"""Dashboard mappings
"""

DASHBOARD_MAPPINGS = {
    "multivector_search": "/sdk/search",
    "cluster_centroids_closest": "/sdk/cluster/centroids/closest",
    "cluster_centroids_furthest": "/sdk/cluster/centroids/furthest",
    "cluster_aggregation": "/sdk/cluster/aggregation",
}
